from collections import OrderedDict

from drf_extra_fields.fields import Base64FieldMixin
from drf_yasg import openapi
from drf_yasg.inspectors import FieldInspector, NotHandled, ViewInspector
from drf_yasg.utils import filter_none, get_serializer_ref_name
from rest_framework import serializers


class FileFieldInspector(FieldInspector):

    def get_schema(self, serializer):
        if self.method not in ViewInspector.body_methods:
            return NotHandled

        # only do this if there are base64 mixin fields
        if any(isinstance(field, Base64FieldMixin) for field in serializer.fields.values()):
            return self.probe_field_inspectors(serializer, openapi.Schema, True)

        return NotHandled

    def field_to_swagger_object(self, field, swagger_object_type, use_references, **kwargs):
        if isinstance(field, serializers.Serializer):
            return self._serializer_to_swagger_object(field, swagger_object_type, use_references, **kwargs)

        if not isinstance(field, Base64FieldMixin):
            return NotHandled

        SwaggerType, ChildSwaggerType = self._get_partial_types(field, swagger_object_type, use_references, **kwargs)

        type_b64 = SwaggerType(type=openapi.TYPE_STRING, format=openapi.FORMAT_BASE64)
        type_uri = SwaggerType(type=openapi.TYPE_STRING, read_only=True, format=openapi.FORMAT_URI)

        if swagger_object_type == openapi.Schema:
            # on writes, it's always b64
            if self.method in ViewInspector.body_methods:
                return type_b64

            # if not representing in base64, it's a link
            return type_uri if not field.represent_in_base64 else type_b64

        return NotHandled

    def _serializer_to_swagger_object(self, serializer, swagger_object_type, use_references, **kwargs):
        if self.method not in ViewInspector.body_methods:
            return NotHandled

        if not any(isinstance(field, Base64FieldMixin) for field in serializer.fields.values()):
            return NotHandled

        SwaggerType, ChildSwaggerType = self._get_partial_types(
            serializer, swagger_object_type,
            use_references, **kwargs
        )

        ref_name = get_serializer_ref_name(serializer)
        ref_name = f"{ref_name}Data" if ref_name else None

        def make_schema_definition():
            properties = OrderedDict()
            required = []
            for property_name, child in serializer.fields.items():
                prop_kwargs = {
                    'read_only': bool(child.read_only) or None
                }
                prop_kwargs = filter_none(prop_kwargs)

                child_schema = self.probe_field_inspectors(
                    child, ChildSwaggerType, use_references, **prop_kwargs
                )
                properties[property_name] = child_schema

                if child.required and not getattr(child_schema, 'read_only', False):
                    required.append(property_name)

            result = SwaggerType(
                type=openapi.TYPE_OBJECT,
                properties=properties,
                required=required or None,
            )
            if not ref_name and 'title' in result:
                # on an inline model, the title is derived from the field name
                # but is visually displayed like the model name, which is confusing
                # it is better to just remove title from inline models
                del result.title

            # Provide an option to add manual paremeters to a schema
            # for example, to add examples
            # self.add_manual_fields(serializer, result)
            return result

        if not ref_name or not use_references:
            return make_schema_definition()

        definitions = self.components.with_scope(openapi.SCHEMA_DEFINITIONS)
        definitions.setdefault(ref_name, make_schema_definition)
        return openapi.SchemaRef(definitions, ref_name)



# Generated by Django 2.0.6 on 2018-11-19 11:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('zds_schema', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='APICredential',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('api_root', models.URLField(unique=True, verbose_name='api root')),
                ('client_id', models.CharField(max_length=255, verbose_name='client id')),
                ('secret', models.CharField(max_length=255, verbose_name='secret')),
            ],
            options={
                'verbose_name': 'API credential',
                'verbose_name_plural': 'API credentials',
            },
        ),
    ]



from .auth import AuthCheckMixin, JWTScopesMixin, generate_jwt
from .schema import TypeCheckMixin, get_operation_url, get_validation_errors

__all__ = [
    'generate_jwt', 'AuthCheckMixin', 'JWTScopesMixin',
    'get_operation_url', 'TypeCheckMixin', 'get_validation_errors'
]



import json
import logging
import os
from urllib.parse import urlsplit

from django.conf import settings

from drf_yasg import openapi
from drf_yasg.codecs import yaml_sane_dump, yaml_sane_load
from drf_yasg.generators import (
    OpenAPISchemaGenerator as _OpenAPISchemaGenerator
)
from drf_yasg.renderers import SwaggerJSONRenderer, SwaggerYAMLRenderer
from drf_yasg.views import get_schema_view
from rest_framework import permissions
from rest_framework.response import Response

logger = logging.getLogger(__name__)


class OpenAPISchemaGenerator(_OpenAPISchemaGenerator):

    def get_path_parameters(self, path, view_cls):
        """Return a list of Parameter instances corresponding to any templated path variables.

        :param str path: templated request path
        :param type view_cls: the view class associated with the path
        :return: path parameters
        :rtype: list[openapi.Parameter]
        """
        parameters = super().get_path_parameters(path, view_cls)

        # see if we can specify UUID a bit more
        for parameter in parameters:
            # the most pragmatic of checks
            if not parameter.name.endswith('_uuid'):
                continue
            parameter.format = openapi.FORMAT_UUID
            parameter.description = "Unieke resource identifier (UUID4)"
        return parameters


DefaultSchemaView = get_schema_view(
    # validators=['flex', 'ssv'],
    generator_class=OpenAPISchemaGenerator,
    public=True,
    permission_classes=(permissions.AllowAny,),
)


class OpenAPIV3RendererMixin:

    def render(self, data, media_type=None, renderer_context=None):
        if 'openapi' in data or 'swagger' in data:
            if self.format == '.yaml':
                return yaml_sane_dump(data, False)
            elif self.format == '.json':
                return json.dumps(data)

        return super().render(data, media_type=media_type, renderer_context=renderer_context)


SPEC_RENDERERS = (
    type(
        'SwaggerYAMLRenderer',
        (OpenAPIV3RendererMixin, SwaggerYAMLRenderer),
        {},
    ),
    type(
        'SwaggerJSONRenderer',
        (OpenAPIV3RendererMixin, SwaggerJSONRenderer),
        {},
    ),
)


class SchemaView(DefaultSchemaView):

    @property
    def _is_openapi_v3(self) -> bool:
        version = self.request.GET.get('v', '')
        return version.startswith('3')

    def get_renderers(self):
        if not self._is_openapi_v3:
            return super().get_renderers()
        return [renderer() for renderer in SPEC_RENDERERS]

    def get(self, request, *args, **kwargs):
        response = super().get(request, *args, **kwargs)

        if not self._is_openapi_v3:
            return response

        # serve the staticically included V3 schema
        SCHEMA_PATH = os.path.join(settings.BASE_DIR, 'src', 'openapi.yaml')
        with open(SCHEMA_PATH, 'r') as infile:
            schema = yaml_sane_load(infile)

        # fix the servers
        for server in schema['servers']:
            split_url = urlsplit(server['url'])
            if split_url.netloc:
                continue
            server['url'] = request.build_absolute_uri(server['url'])

        return Response(
            data=schema,
            headers={'X-OAS-Version': schema['openapi']}
        )



from rest_framework import serializers

from ..models import JWTSecret


class JWTSecretSerializer(serializers.ModelSerializer):
    class Meta:
        model = JWTSecret
        fields = ('identifier', 'secret')



import time

import jwt
from rest_framework import status

from ..models import JWTSecret
from ..scopes import Scope


def generate_jwt(scopes: list, secret: str='letmein', zaaktypes: list=None) -> str:
    scope_labels = sum((_get_scope_labels(scope) for scope in scopes), [])
    payload = {
        # standard claims
        'iss': 'testsuite',
        'iat': int(time.time()),
        # custom claims
        'zds': {
            'scopes': scope_labels,
            'zaaktypes': zaaktypes or [],
        },
    }
    headers = {
        'client_identifier': 'testsuite',
    }
    encoded = jwt.encode(payload, secret, headers=headers, algorithm='HS256')
    encoded = encoded.decode('ascii')
    return f"Bearer {encoded}"


def _get_scope_labels(scope) -> list:
    if not scope.children:
        return [scope.label]

    labels = []
    for child in scope.children:
        labels += _get_scope_labels(child)
    return sorted(set(labels))


class AuthCheckMixin:

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        JWTSecret.objects.get_or_create(
            identifier='testsuite',
            defaults={'secret': 'letmein'}
        )

    def assertForbidden(self, url, method='get'):
        """
        Assert that an appropriate scope is required.
        """
        do_request = getattr(self.client, method)

        with self.subTest(case='JWT missing'):
            response = do_request(url)

            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        with self.subTest(case='Invalid JWT structure'):
            invalid_jwt = generate_jwt(scopes=[Scope('invalid.scope')])[:-10]
            self.client.credentials(HTTP_AUTHORIZATION=invalid_jwt)

            response = do_request(url)

            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        with self.subTest(case='Correct scope missing'):
            jwt = generate_jwt(scopes=[Scope('invalid.scope')])
            self.client.credentials(HTTP_AUTHORIZATION=jwt)

            response = do_request(url)

            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def assertForbiddenWithCorrectScope(
            self, url: str, scopes: list, method='get',
            request_kwargs=None, **extra_claims):

        do_request = getattr(self.client, method)
        request_kwargs = request_kwargs or {}

        jwt = generate_jwt(scopes=scopes, **extra_claims)
        self.client.credentials(HTTP_AUTHORIZATION=jwt)

        response = do_request(url, **request_kwargs)

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)


class JWTScopesMixin:

    scopes = None
    zaaktypes = None

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        JWTSecret.objects.get_or_create(
            identifier='testsuite',
            defaults={'secret': 'letmein'}
        )

    def setUp(self):
        super().setUp()

        if self.scopes is not None:
            token = generate_jwt(
                scopes=self.scopes,
                zaaktypes=self.zaaktypes or [],
                secret='letmein'
            )
            self.client.credentials(HTTP_AUTHORIZATION=token)



from django.utils.decorators import method_decorator

from drf_yasg.utils import swagger_auto_schema


def action_description(action: str, description: str):
    decorator = swagger_auto_schema(operation_description=description)
    return method_decorator(name=action, decorator=decorator)
